/**
        qled - Copyright (C) 2006, 2007 Simeon Simeonov

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software Foundation,
 Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

**/


#include <QtGui>
#include <QtOpenGL>


#include "glwidget.h"

/* coordinates */
#define X_LOWER -1.0
#define X_UPPER 1.0

#define Y_LOWER -1.0
#define Y_UPPER 1.0

#define Z_LOWER -1.0
#define Z_UPPER 1.0


/* options masks */
#define SHOW_DEC_TIME 0x1
#define SHOW_AMPM 0x2
#define DISPLAY_HELP 0x4


GLWidget::GLWidget(QWidget *parent) : QGLWidget(parent)
{
  
  d_object = 0;
  l_object = 0;

  rangle = 0;
  settings = 0x0;

  settingsDisable(SHOW_DEC_TIME | SHOW_AMPM | DISPLAY_HELP);

  QTimer *timer = new QTimer(this);
  connect(timer, SIGNAL(timeout()), this, SLOT(executeTicks()));
  timer->start(200);
    
}


GLWidget::~GLWidget()
{

  makeCurrent();
  glDeleteLists(d_object, 1);
  glDeleteLists(l_object, 1);
    
}


QSize 
GLWidget::minimumSizeHint() const
{
  return QSize(80, 60);
}


QSize 
GLWidget::sizeHint() const
{
  return QSize(800, 600);
}


void GLWidget::initializeGL()
{

  /* set the clear-values for the color and the depth buffers */
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClearDepth(1.0);

  glShadeModel(GL_SMOOTH);
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_DEPTH_TEST);
  
    /* Light stuff */
  GLfloat lmodel_ambient[] = {0.4, 0.4, 0.4, 1.0};
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
  glEnable(GL_LIGHTING);

  l_object = makeLObject();
  d_object = makeDObject();

  /* "old" values set to current time in the beginning */
  /* get the current time */
  minutes = old_minutes = QTime::currentTime().minute();
  hours = old_hours = QTime::currentTime().hour();

  /* the "global" x and y rotation angles are set to 0 */
  x_rotation = y_rotation = 0;

  /* no need to consider back-facing polygons */
  glEnable(GL_CULL_FACE);

}


void GLWidget::paintGL()
{

  int i;
  int seconds; /* no need to make seconds a class member */
  int tmp_h; /* stores temporary the value of old_hours in case we are using the AM/PM format */


  int h_led[5];
  int m_led[6];
  int s_led[6];

  int h_changed[5];
  int m_changed[6];

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  glPushMatrix();
  gluLookAt(0.0, 0.0, 20.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

  /* Time calculation */
  QTime t = QTime::currentTime(); /* get the current time */
  minutes = t.minute();
  hours = t.hour();
  seconds = t.second();


  /* global rotation */
  glRotated((GLdouble) x_rotation, 1.0, 0.0, 0.0);
  glRotated((GLdouble) y_rotation, 0.0, 1.0, 0.0);
  /* --------------- */

  /* display help */
  if (settings & DISPLAY_HELP)
    displayHelp();


  /* display the decimal time */
  if (settings & SHOW_DEC_TIME) {
    QFont serifFont("Times", 16);

    if (settings & SHOW_AMPM)
      renderText (10.5, -9.0, 1.0, t.toString("hh:mm:ss ap"), serifFont);
    else
      renderText (10.5, -9.0, 1.0, t.toString("hh:mm:ss"), serifFont);

  }


  /* perform calculations */

  /* hour calculations */
  if (hours != old_hours) {

    if (settings & SHOW_AMPM) {

      if (hours > 12) {
        hours -= 12;
        h_changed[4] = 1;
      } else {
        h_changed[4] = 0;
      }

      tmp_h = old_hours;
      if (tmp_h > 12) {
        tmp_h = old_hours - 12;
        h_led[4] = 1;
      } else {
        h_led[4] = 0;
      }
      
      getLedArray(h_led, tmp_h, 4);
      getLedArray(h_changed, hours, 4);

    } else {

     getLedArray(h_led, old_hours, 5);
     getLedArray(h_changed, hours, 5);

    }    

  } else {

    if (settings & SHOW_AMPM) {

      tmp_h = hours;
      if (tmp_h > 12) {
        tmp_h -= 12;
        h_led[4] = 1;
      } else {
        h_led[4] = 0;
      }

      getLedArray(h_led, tmp_h, 4);

    } else {

      getLedArray(h_led, hours, 5);

    }

  }


  if (minutes != old_minutes) {

    getLedArray(m_led, old_minutes, 6);
    getLedArray(m_changed, minutes, 6);
 
  } else {
  
    getLedArray(m_led, minutes, 6);

  }

  getLedArray(s_led, seconds, 6);



  /* Now we start the drawing routines */

  /* hours */

  for (i = -2; i < 3; i++) {

    glPushMatrix();

    glTranslated(-1.0 + i * 4.0, 8.0, 0.0);
    if ( (hours != old_hours) && (h_led[i+2] != h_changed[i+2]))
      if (h_changed[i+2])
        glRotated( (GLdouble) rangle, 1.0, 0.0, 0.0); /* rotate CCW if the led is changing from 0 to 1 */
      else 
        glRotated( - (GLdouble) rangle, 1.0, 0.0, 0.0); /* rotate CW if the led is changing from 1 to 0 */

    if ( (settings & SHOW_AMPM) && (i == 2))
      glScaled(0.5, 0.5, 0.5);

    if (h_led[i+2])
      glCallList(l_object);
    else
      glCallList(d_object);
  
    glPopMatrix();
  
  }


  /* Minutes */

  for (i = -2; i < 4; i++) {

    glPushMatrix();

    glTranslated(-3.0 + i * 4.0, 0.0, 0.0);
    if ( (minutes != old_minutes) && (m_led[i+2] != m_changed[i+2]))
      if (m_changed[i+2])
        glRotated( (GLdouble) rangle, 1.0, 0.0, 0.0); /* rotate CCW if the led is changing from 0 to 1 */
      else
        glRotated( - (GLdouble) rangle, 1.0, 0.0, 0.0); /* rotate CW if the led is changing from 1 to 0 */
    if (m_led[i+2])
      glCallList(l_object);  
    else
      glCallList(d_object);
  
    glPopMatrix();

  } 


  /* Seconds */

  
  for (i = -2; i < 4; i++) {

    glPushMatrix();

    glTranslated(-3.0 + i * 4.0, -8.0, 0.0);
    if (s_led[i+2])
      glCallList(l_object);  
    else
      glCallList(d_object);
  
    glPopMatrix();
    
  }

  glColor4f(1.0, 1.0, 1.0, 1.0);

  glPopMatrix();
  swapBuffers();
  
}


void GLWidget::resizeGL(int width, int height)
{

  GLfloat light_position[] = {0.0, 0.0, 0.0, 1.0};
  glViewport(0, 0, width, height);
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(60.0, (GLdouble) width / (GLdouble) height, 5.0, 60.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);
  
}


void
GLWidget::keyPressEvent (QKeyEvent * event)
{

  switch (event->key()) {

  case Qt::Key_Q:
    close();
    break;
    
  case Qt::Key_L:
    if (glIsEnabled(GL_LIGHT0))
      glDisable(GL_LIGHT0);
    else 
      glEnable(GL_LIGHT0);
    break;

  case Qt::Key_D:
    if (settings & SHOW_DEC_TIME)
      settingsDisable(SHOW_DEC_TIME);
    else 
      settingsEnable(SHOW_DEC_TIME);
    break;
   
  case Qt::Key_A:
    if (settings & SHOW_AMPM) {
      settingsDisable(SHOW_AMPM);
      glDisable(GL_RESCALE_NORMAL); /* no scaling will be performed, so there is no need to waste resources */
    } else {
      settingsEnable(SHOW_AMPM);
      glEnable(GL_RESCALE_NORMAL); /* the last hour-led will be uniformly scaled */
    }
    break;

  case Qt::Key_F:
    if (isFullScreen()) 
      showMaximized();
    else
      showFullScreen();
    updateGL();
    break;

  case Qt::Key_H:
    if (settings & DISPLAY_HELP)
      settingsDisable(DISPLAY_HELP);
    else
      settingsEnable(DISPLAY_HELP);
    updateGL();
    break;

  case Qt::Key_R:
    x_rotation = y_rotation = 0;
    updateGL();
    break;

  }


}


void 
GLWidget::mousePressEvent(QMouseEvent *event)
{

 if (event->buttons() & Qt::RightButton) {

    x_rotation = y_rotation = 0;
    updateGL();

  }

  old_pos = event->pos();

}


void 
GLWidget::mouseDoubleClickEvent(QMouseEvent *event)
{

  if (event->buttons() & Qt::LeftButton) {
    if (isFullScreen())
      showMaximized();
    else
      showFullScreen();
    updateGL();  
  }

}


void 
GLWidget::mouseMoveEvent(QMouseEvent *event)
{

  int dx = event->x() - old_pos.x();
  int dy = event->y() - old_pos.y();

  if (event->buttons() & Qt::LeftButton) {
 
   x_rotation += dy >> 3;
   y_rotation += dx >> 3;

    if ( (x_rotation > 360) || (x_rotation < -360))
      x_rotation = x_rotation % 360;

    if ( (y_rotation > 360) || (y_rotation < -360))
      y_rotation = y_rotation % 360;

    updateGL();

  }

  old_pos = event->pos();

}


void GLWidget::executeTicks()
{

  if (minutes != old_minutes) {
    rangle += 5;
    if (rangle >= 90) { /* the 90-degree turn has been made. turn off the animation */
      rangle = 0;
      old_minutes = QTime::currentTime().minute();
      old_hours = QTime::currentTime().hour();
    }

  }

  updateGL();

}


GLuint GLWidget::makeLObject() const
{

  GLuint list = glGenLists(1);
  glNewList(list, GL_COMPILE);
  
  glColor4f(1.0, 0.0, 0.0, 1.0);

  GLfloat mat_emission[] = {0.7, 0.0, 0.0, 0.0};
  glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, mat_emission);


  glBegin(GL_QUADS); /* Back side */
  glNormal3f(0.0, 0.0, -1.0);
  glVertex3f(X_LOWER, Y_LOWER, Z_LOWER);
  glVertex3f(X_LOWER, Y_UPPER, Z_LOWER);
  glVertex3f(X_UPPER, Y_UPPER, Z_LOWER);
  glVertex3f(X_UPPER, Y_LOWER, Z_LOWER);
  glEnd();

  glBegin(GL_QUADS); /* Front side */
  glNormal3f(0.0, 0.0, 1.0);
  glVertex3f(X_LOWER, Y_LOWER, Z_UPPER);
  glVertex3f(X_UPPER, Y_LOWER, Z_UPPER);
  glVertex3f(X_UPPER, Y_UPPER, Z_UPPER);
  glVertex3f(X_LOWER, Y_UPPER, Z_UPPER);
  glEnd();

  glBegin(GL_QUADS); /* Bottom side */
  glNormal3f(0.0, -1.0, 0.0);
  glVertex3f(X_LOWER, Y_LOWER, Z_LOWER);
  glVertex3f(X_UPPER, Y_LOWER, Z_LOWER);
  glVertex3f(X_UPPER, Y_LOWER, Z_UPPER);
  glVertex3f(X_LOWER, Y_LOWER, Z_UPPER);
  glEnd();

  glBegin(GL_QUADS); /* Top side */
  glNormal3f(0.0, 1.0, 0.0);
  glVertex3f(X_LOWER, Y_UPPER, Z_LOWER);
  glVertex3f(X_LOWER, Y_UPPER, Z_UPPER);
  glVertex3f(X_UPPER, Y_UPPER, Z_UPPER);
  glVertex3f(X_UPPER, Y_UPPER, Z_LOWER);
  glEnd();

  glBegin(GL_QUADS); /* Left side */
  glNormal3f(-1.0, 0.0, 0.0);
  glVertex3f(X_LOWER, Y_LOWER, Z_LOWER);
  glVertex3f(X_LOWER, Y_LOWER, Z_UPPER);
  glVertex3f(X_LOWER, Y_UPPER, Z_UPPER);
  glVertex3f(X_LOWER, Y_UPPER, Z_LOWER);
  glEnd();

  glBegin(GL_QUADS); /* Right side */
  glNormal3f(1.0, 0.0, 0.0);
  glVertex3f(X_UPPER, Y_LOWER, Z_LOWER);
  glVertex3f(X_UPPER, Y_UPPER, Z_LOWER);
  glVertex3f(X_UPPER, Y_UPPER, Z_UPPER);
  glVertex3f(X_UPPER, Y_LOWER, Z_UPPER);
  glEnd();


  glEndList();
  return list;

}


GLuint GLWidget::makeDObject() const
{

  GLuint list = glGenLists(1);
  glNewList(list, GL_COMPILE);
  
  glColor4f(1.0, 0.0, 0.0, 1.0);
  
  GLfloat mat_emission[] = {0.0, 0.0, 0.0, 0.0};
  glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, mat_emission);


  glBegin(GL_QUADS); /* Back side */
  glNormal3f(0.0, 0.0, -1.0);
  glVertex3f(X_LOWER, Y_LOWER, Z_LOWER);
  glVertex3f(X_LOWER, Y_UPPER, Z_LOWER);
  glVertex3f(X_UPPER, Y_UPPER, Z_LOWER);
  glVertex3f(X_UPPER, Y_LOWER, Z_LOWER);
  glEnd();

  glBegin(GL_QUADS); /* Front side */
  glNormal3f(0.0, 0.0, 1.0);
  glVertex3f(X_LOWER, Y_LOWER, Z_UPPER);
  glVertex3f(X_UPPER, Y_LOWER, Z_UPPER);
  glVertex3f(X_UPPER, Y_UPPER, Z_UPPER);
  glVertex3f(X_LOWER, Y_UPPER, Z_UPPER);
  glEnd();

  glBegin(GL_QUADS); /* Bottom side */
  glNormal3f(0.0, -1.0, 0.0);
  glVertex3f(X_LOWER, Y_LOWER, Z_LOWER);
  glVertex3f(X_UPPER, Y_LOWER, Z_LOWER);
  glVertex3f(X_UPPER, Y_LOWER, Z_UPPER);
  glVertex3f(X_LOWER, Y_LOWER, Z_UPPER);
  glEnd();

  glBegin(GL_QUADS); /* Top side */
  glNormal3f(0.0, 1.0, 0.0);
  glVertex3f(X_LOWER, Y_UPPER, Z_LOWER);
  glVertex3f(X_LOWER, Y_UPPER, Z_UPPER);
  glVertex3f(X_UPPER, Y_UPPER, Z_UPPER);
  glVertex3f(X_UPPER, Y_UPPER, Z_LOWER);
  glEnd();

  glBegin(GL_QUADS); /* Left side */
  glNormal3f(-1.0, 0.0, 0.0);
  glVertex3f(X_LOWER, Y_LOWER, Z_LOWER);
  glVertex3f(X_LOWER, Y_LOWER, Z_UPPER);
  glVertex3f(X_LOWER, Y_UPPER, Z_UPPER);
  glVertex3f(X_LOWER, Y_UPPER, Z_LOWER);
  glEnd();

  glBegin(GL_QUADS); /* Right side */
  glNormal3f(1.0, 0.0, 0.0);
  glVertex3f(X_UPPER, Y_LOWER, Z_LOWER);
  glVertex3f(X_UPPER, Y_UPPER, Z_LOWER);
  glVertex3f(X_UPPER, Y_UPPER, Z_UPPER);
  glVertex3f(X_UPPER, Y_LOWER, Z_UPPER);
  glEnd();


  glEndList();
  return list;

}


void
GLWidget::getLedArray(int *array, int value, int size)
{

  int i;
 
  for (i = size - 1; i >= 0; i--) {
    
    if (value%2)
      array[i] = 1;
    else
      array[i] = 0;
    
    value >>= 1;
    
  }

}


void 
GLWidget::displayHelp()
{

  QFont serifFont("Times", 14);
  renderText (-13.0, 8.0, 3.0, "qled-1.0", serifFont);
  renderText (-13.0, 6.0, 3.0, "Keyboard:", serifFont);
  renderText (-12.0, 5.0, 3.0, "a - toggle between European (default) and AM/PM time format", serifFont);
  renderText (-12.0, 4.0, 3.0, "d - display decimal clock on / off (default)", serifFont);
  renderText (-12.0, 3.0, 3.0, "f - toggle full screen (default - off)", serifFont);
  renderText (-12.0, 2.0, 3.0, "h - display this message (default - off)", serifFont);
  renderText (-12.0, 1.0, 3.0, "l - toggle stationary light source (default - off)", serifFont);
  renderText (-12.0, 0.0, 3.0, "q - quit", serifFont);
  renderText (-12.0, -1.0, 3.0, "r - reset the clock's position", serifFont);
  renderText (-13.0, -3.0, 3.0, "Mouse: ", serifFont);
  renderText (-12.0, -4.0, 3.0, "double click (left button) - toggle full screen (default - off)", serifFont);
  renderText (-12.0, -5.0, 3.0, "left click + drag - rotate the clock", serifFont);
  renderText (-12.0, -6.0, 3.0, "right click - reset the clock's position", serifFont);

}
